// Copyright (C) 2013 Garth N. Wells
//
// This file is part of DOLFIN.
//
// DOLFIN is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// DOLFIN is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with DOLFIN. If not, see <http://www.gnu.org/licenses/>.
//
// First added:  2013-01-10
// Last changed:

#include <limits>
#include <map>
#include <utility>
#include <vector>
#include <boost/unordered_map.hpp>

#include <dolfin/common/Array.h>
#include <dolfin/log/log.h>
#include <dolfin/mesh/BoundaryMesh.h>
#include <dolfin/mesh/Cell.h>
#include <dolfin/mesh/Facet.h>
#include <dolfin/mesh/Mesh.h>
#include <dolfin/mesh/MeshEntityIterator.h>
#include <dolfin/mesh/SubDomain.h>
#include <dolfin/mesh/Vertex.h>
#include "PeriodicBoundaryComputation.h"

using namespace dolfin;

// Comparison operator for hashing coordinates. Note that two
// coordinates are considered equal if equal to within round-off.
struct lt_coordinate
{
  bool operator() (const std::vector<double>& x,
                   const std::vector<double>& y) const
  {
    std::size_t n = std::max(x.size(), y.size());
    for (std::size_t i = 0; i < n; ++i)
    {
      double xx = 0.0;
      double yy = 0.0;
      if (i < x.size())
        xx = x[i];
      if (i < y.size())
        yy = y[i];

      if (xx < (yy - DOLFIN_EPS))
        return true;
      else if (xx > (yy + DOLFIN_EPS))
        return false;
    }
    return false;
  }
};

//-----------------------------------------------------------------------------
std::map<std::size_t, std::pair<std::size_t, std::size_t> >
  PeriodicBoundaryComputation::compute_periodic_pairs(const Mesh& mesh,
                                                const SubDomain& sub_domain,
                                                const std::size_t dim)
{
  // Get geometric dimension
  const std::size_t gdim = mesh.geometry().dim();

  // Create boundary mesh
  BoundaryMesh bmesh(mesh);

  // Arrays used for mapping coordinates
  std::vector<double> x(gdim);
  std::vector<double> y(gdim);

  // Wrap x and y (Array view of x and y)
  Array<double> _x(gdim, x.data());
  Array<double> _y(gdim, y.data());

  std::vector<std::size_t> slave_entities;
  std::vector<std::vector<double> > slave_mapped_coords;

  // Min/max coordinates of facet midpoints [min_x, max_x]. Used to
  // build bounding box of all master entity midpoints on this process
  std::vector<double> x_min_max;

  // Map from master entity midpoint coordinate to local facet index
  std::map<std::vector<double>, std::size_t, lt_coordinate> master_coord_to_entity_index;

  // Iterate over entities to find master/slave facets
  const MeshFunction<std::size_t>& entity_map = bmesh.entity_map(dim);
  for (MeshEntityIterator e(bmesh, dim); !e.end(); ++e)
  {
    // Create entity
    const MeshEntity entity(mesh, dim, entity_map[*e]);

    // Copy entity coordinate
    const Point midpoint = entity.midpoint();
    std::copy(midpoint.coordinates(), midpoint.coordinates() + gdim,
              x.begin());

    // Check if entity lies on a 'master' or 'slave' boundary
    if (sub_domain.inside(_x, true))
    {
      // Build bounding box data for master entity midpoints
      if (x_min_max.empty())
      {
        x_min_max = x;
        x_min_max.insert(x_min_max.end(), x.begin(), x.end());
      }
      for (std::size_t i = 0; i < gdim; ++i)
      {
       x_min_max[i]        = std::min(x_min_max[i], x[i]);
       x_min_max[i + gdim] = std::max(x_min_max[i + gdim], x[i]);
      }

      // Insert (midpoint coordinates, local index) into map
      master_coord_to_entity_index.insert(std::make_pair(x, entity.index()));
    }
    else
    {
      // Get mapped midpoint (y) of slave entity
      sub_domain.map(_x, _y);

      // Check if entity lies on a 'slave' boundary
      if (sub_domain.inside(_y, true))
      {
        // Store slave local index and midpoint coordinates
        slave_entities.push_back(entity.index());
        slave_mapped_coords.push_back(y);
      }

    }
  }

  // Communicate bounding boxes for master entities
  std::vector<std::vector<double> > bounding_boxes;
  MPI::all_gather(x_min_max, bounding_boxes);

  // Number of MPI processes
  std::size_t num_processes = MPI::num_processes();

  // Build send buffer of mapped slave midpoint coordinate to processes
  // that may own the master entity
  std::vector<std::vector<double> > slave_mapped_coords_send(num_processes);
  std::vector<std::vector<std::size_t> > sent_slave_indices(num_processes);
  for (std::size_t i = 0; i < slave_entities.size(); ++i)
  {
    for (std::size_t p = 0; p < num_processes; ++p)
    {
      // Slave mapped coordinates from process p
      std::vector<double>& slave_mapped_coords_send_p = slave_mapped_coords_send[p];

      // Check if mapped slave falls within master entity bounding box
      // on process p
      if (in_bounding_box(slave_mapped_coords[i], bounding_boxes[p]))
      {
        sent_slave_indices[p].push_back(slave_entities[i]);
        slave_mapped_coords_send_p.insert(slave_mapped_coords_send_p.end(),
                                          slave_mapped_coords[i].begin(),
                                          slave_mapped_coords[i].end());
      }
    }
  }

  // Send slave midpoints to possible owners of correspoding master entity
  std::vector<std::vector<double> > slave_mapped_coords_recv;
  MPI::all_to_all(slave_mapped_coords_send, slave_mapped_coords_recv);
  dolfin_assert(slave_mapped_coords_recv.size() == num_processes);

  // Check if this process owns the master facet for a reveived (mapped)
  std::vector<double> coordinates(gdim);
  std::vector<std::vector<std::size_t> > master_local_entity(num_processes);
  for (std::size_t p = 0; p < num_processes; ++p)
  {
    const std::vector<double>& slave_mapped_coords_p = slave_mapped_coords_recv[p];
    for (std::size_t i = 0; i < slave_mapped_coords_p.size(); i += gdim)
    {
      // Unpack received mapped slave midpoint coordinate
      std::copy(&slave_mapped_coords_p[i],
                &slave_mapped_coords_p[i] + gdim, coordinates.begin());

      // Check is this process has a master entity that is paired with
      // a received slave entity
      std::map<std::vector<double>, std::size_t, lt_coordinate>::const_iterator
        it = master_coord_to_entity_index.find(coordinates);

      // If this process owns the master, insert master entity index,
      // else insert std::numeric_limits<std::size_t>::max()
      if (it !=  master_coord_to_entity_index.end())
        master_local_entity[p].push_back(it->second);
      else
        master_local_entity[p].push_back(std::numeric_limits<std::size_t>::max());
    }
  }

  // Send local index of master entity back to owner of slave entity
  std::vector<std::vector<std::size_t> > master_entity_local_index_recv;
  MPI::all_to_all(master_local_entity,  master_entity_local_index_recv);

  // Build map from slave facets on this process to master facet (local
  // facet index, process owner)
  std::map<std::size_t, std::pair<std::size_t, std::size_t> > slave_to_master_entity;
  std::size_t num_local_slave_entities = 0;
  for (std::size_t p = 0; p < num_processes; ++p)
  {
    const std::vector<std::size_t> master_entity_index_p = master_entity_local_index_recv[p];
    const std::vector<std::size_t> sent_slaves_p = sent_slave_indices[p];
    dolfin_assert(master_entity_index_p.size() == sent_slaves_p.size());

    for (std::size_t i = 0; i < master_entity_index_p.size(); ++i)
    {
      if (master_entity_index_p[i] < std::numeric_limits<std::size_t>::max())
      {
        ++num_local_slave_entities;
        if (!slave_to_master_entity.insert(std::make_pair(sent_slaves_p[i],
                                     std::make_pair(p, master_entity_index_p[i]))).second)
        {
          dolfin_error("PeriodicDomain.cpp",
                       "build peridic master-slave mapping",
                       "More than one master entity for slave entity");
        }
      }
    }
  }

  // Number global master and slave facets
  const std::size_t num_global_master_entities = MPI::sum(master_coord_to_entity_index.size());
  const std::size_t num_global_slave_entities = MPI::sum(num_local_slave_entities);

  // Check that number of global master and slave entities match
  if (num_global_master_entities != num_global_slave_entities)
  {
    dolfin_error("PeriodicDomain.cpp",
                 "global number of slave and master facets",
                 "Number of slave and master facets is not equal");
  }

  return slave_to_master_entity;
}
//-----------------------------------------------------------------------------
bool PeriodicBoundaryComputation::in_bounding_box(const std::vector<double>& point,
                                     const std::vector<double>& bounding_box)
{
  // Return false if bounding box is empty
  if (bounding_box.empty())
    return false;

  const std::size_t gdim = point.size();
  dolfin_assert(bounding_box.size() == 2*gdim);
  for (std::size_t i = 0; i < gdim; ++i)
  {
    if (!(point[i] >= (bounding_box[i] - DOLFIN_EPS)
      && point[i] <= (bounding_box[gdim + i] + DOLFIN_EPS)))
    {
      return false;
    }
  }
  return true;
}
//-----------------------------------------------------------------------------
